structure Translate = 
struct
  (*  fresh () ==> a string which has not previously been returned by this
   *  function, and which could not have appeared in the source program.
   *)
  local
    val c = ref 0
  in
    fun fresh () = (
      c := !c + 1;
      "%" ^ Int.toString (!c)
    )
  end

  structure LPP = LambdaPlusPlus
  structure LC = LambdaCalculus

  val ` = LC.Variable
  val `` = LPP.Variable 

  (* a few popular combinators in LC, for convenience *)
  local val (x, y) = (fresh (), fresh ()) in
    val K = LC.Lambda (x, LC.Lambda (y, `x))
    val I = LC.Lambda (x, `x)
    val K' = LC.Lambda (x, I)
    val Y = LC.Lambda (x, LC.Apply (
      LC.Lambda (y, LC.Apply (`x, LC.Apply (`y, `y))),
      LC.Lambda (y, LC.Apply (`x, LC.Apply (`y, `y)))
    ))
  end


  val rec translate = (fn
    LPP.Variable x => `x
  | LPP.Lambda (x, e) => LC.Lambda (x, translate e)
  | LPP.Apply (e1, e2) => LC.Apply (translate e1, translate e2)
  | LPP.Tuple (e1, e2) =>
    let val x = fresh () in
      LC.Lambda (
        x,
        LC.Apply (LC.Apply (`x, translate e1), translate e2)
      )
    end
  | LPP.First e => LC.Apply (translate e, K)
  | LPP.Second e => LC.Apply (translate e, K')
  | LPP.Left e =>
    let val (x, y) = (fresh (), fresh ()) in
      LC.Lambda (x, LC.Lambda (y, LC.Apply (`x, translate e)))
    end
  | LPP.Right e =>
    let val (x, y) = (fresh (), fresh ()) in
      LC.Lambda (x, LC.Lambda (y, LC.Apply (`y, translate e)))
    end
  | LPP.Case (e, (x1, e1), (x2, e2)) =>
      LC.Apply (
        LC.Apply (
          translate e,
          LC.Lambda (x1, translate e1)
        ),
        LC.Lambda (x2, translate e2)
      )
  | LPP.Fix (x, e) => LC.Apply (Y, LC.Lambda (x, translate e))
  )
end